/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef BEN_BUFFER_H
#define BEN_BUFFER_H
#include <stdarg.h>
#include <unistd.h>

typedef struct buffer_t {
	size_t mask, size, alloc, partial;
	void *data;
	int eof, error;
	size_t total_read, total_write;
} buffer;


void buffer_init(buffer *b, size_t gran);
void buffer_clear(buffer *b);
void buffer_reset(buffer *b);

void buffer_alloc_nt(buffer *b, size_t size);

static inline void buffer_alloc(buffer *b, size_t size)
{
	if (size > b->alloc)
		buffer_alloc_nt(b, size);
}

static inline void *buffer_get(buffer *b, size_t size)
{
	buffer_alloc(b, size);
	b->size = size;
	return(b->data);
}

static inline void buffer_memcpy(buffer *b, const void *src, size_t size)
{
	buffer_get(b, size);
	__builtin_memcpy(b->data, src, size);
}

static inline void buffer_memappend(buffer *b, const void *src, size_t size)
{
	size_t o;
	
	o = b->size;
	buffer_get(b, o + size);
	
	__builtin_memcpy(b->data + o, src, size);
}

static inline void buffer_strcpy(buffer *b, const char *src)
{
	size_t len;
	
	len = __builtin_strlen(src);
	buffer_memcpy(b, src, len);
}

static inline void buffer_strappend(buffer *b, const char *src)
{
	size_t len;
	
	len = __builtin_strlen(src);
	buffer_memappend(b, src, len);
}

char *buffer_str(buffer *b);
char *buffer_fappend(buffer *b, const char *format, ...);

int buffer_load(buffer *b, int fd, size_t max_size);
int buffer_save(buffer *b, int fd);
int buffer_load_path(buffer *b, const char *path, size_t max_size);
int buffer_save_path(buffer *b, const char *path);

void buffer_io_reset(buffer *b);
int buffer_io_read(buffer *b, int fd);
int buffer_io_write(buffer *b, int fd);
int buffer_io_complete(buffer *b);
int buffer_io_eof(buffer *b);
int buffer_io_error(buffer *b);

#endif

