/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <poll.h>
#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "xmem.h"
#include "network.h"
#include "block.h"
#include "date.h"
#include "protocol.h"
#include "common.h"
#include "config.h"
#include "buffer.h"
#include "ini.h"
#include "exec.h"
#include "node.h"
#include "control.h"

// **************************************************************************
// 
// **************************************************************************
#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

#define CONTROL_SEND_ONLY	1

typedef struct odir_info_t {
	char name[32];
	uint64_t start, stop;
	int is_exec, done, running, pending;
} odir_info;

typedef struct control_t {
	const char *action;
	int fd;
	int flags;
	buffer inbuff;
	buffer outbuff;
	
	uint64_t g_start, g_stop;
	int g_done, g_running, g_pending;
	int nodir;
	buffer odir;
} control;

// **************************************************************************
// 
// **************************************************************************
static int control_init(control *k, char *action,
	char *addr)
{
	char *name;
	
	k->action = action;

	k->fd = socket_connect(addr);
	
	if (k->fd < 0)
		return(-1);
	
	buffer_init(&k->inbuff, 0);
	buffer_init(&k->outbuff, 0);
	
	k->g_start = 0;
	k->g_stop = 0;
	k->g_done = k->g_running = k->g_pending = 0;
	
	k->nodir = 0;
	buffer_init(&k->odir, 0);

	name = xprintf(NULL, "%s: %s", conf.host, action);
	
	proto_append(&k->outbuff, PROTO_CL_MAGIC, PROTO_CL_SETUP,
		"is", -1, name);

	xfree(name);
	
	return(0);
}

static void control_clear(control *k)
{
	close(k->fd);
	buffer_clear(&k->inbuff);
	buffer_clear(&k->outbuff);
	buffer_clear(&k->odir);
}

// **************************************************************************
// 
// **************************************************************************
static void control_gather_job(control *k, char *tdir, int type, int is_exec,
	uint64_t start, uint64_t stop)
{
	odir_info *odir;
	int n, o;
	
	odir = (odir_info *)k->odir.data;
	n = k->nodir;
	
	for (o = 0; o < n; o++) {
		if ((odir[o].is_exec == is_exec)
		&&  (strcmp(odir[o].name, tdir) == 0))
			break;
	}
	
	if (o == n) {
		odir_info d;
		
		strncpy(d.name, tdir, sizeof(d.name));
		d.name[sizeof(d.name) - 1] = 0;
		
		d.start = 0;
		d.stop = 0;
		d.is_exec = is_exec;
		d.done = 0;
		d.running = 0;
		d.pending = 0;
		
		buffer_memappend(&k->odir, &d, sizeof(odir_info));
		odir = (odir_info *)k->odir.data;
		
		k->nodir++;
	}
	
	if (type == PROTO_TASK_DONE) {
		k->g_done++;
		odir[o].done++;
	} else if (type == PROTO_TASK_RUNNING) {
		k->g_running++;
		odir[o].running++;
	} else if (type == PROTO_TASK_PENDING) {
		k->g_pending++;
		odir[o].pending++;
	}
	
	if (type != PROTO_TASK_PENDING) {
		if ((k->g_start == 0) || (start < k->g_start))
			k->g_start = start;
		if (stop > k->g_stop)
			k->g_stop = stop;

		if ((odir[o].start == 0) || (start < odir[o].start))
			odir[o].start = start;
		if (stop > odir[o].stop)
			odir[o].stop = stop;
	}
}

// **************************************************************************
// 
// **************************************************************************
static void control_print(int h, int tid, int cid, char *host,
	char *d, size_t sz)
{
	FILE *f;
	size_t o, z;
	
	f = (h == PROTO_SV_OUT) ? stdout : stderr;
	
	o = 0;
	while (o < sz) {
		z = o;
		while ((z < sz) && (d[z] != '\n'))
			z++;
		
		fprintf(f, "[%d](%d %s): ", tid, cid, host);
		fwrite(d + o, z - o, 1, f);
		fprintf(f, "%s\n", (z < sz) ? "" : "[...]");
		
		o = z + 1;
	}
}

// **************************************************************************
// 
// **************************************************************************
enum tag_t {
	TAG_JOB_HEADER_BEGIN,
	TAG_JOB_HEADER_END,
	TAG_JOB_ROW_BEGIN,
	TAG_JOB_ROW_END,
	TAG_JOB_EXEC_BEGIN,
	TAG_JOB_EXEC_END,
	TAG_JOB_REGULAR_BEGIN,
	TAG_JOB_REGULAR_END,
	TAG_JOB_DONE_BEGIN,
	TAG_JOB_DONE_END,
	TAG_JOB_RUNNING_BEGIN,
	TAG_JOB_RUNNING_END,
	TAG_JOB_PENDING_BEGIN,
	TAG_JOB_PENDING_END,
	TAG_NODE_HEADER_BEGIN,
	TAG_NODE_HEADER_END,
	TAG_NODE_ROW_BEGIN,
	TAG_NODE_ROW_END,
	TAG_NODE_CONTROL_BEGIN,
	TAG_NODE_CONTROL_END,
	TAG_NODE_COMPUTE_BEGIN,
	TAG_NODE_COMPUTE_END,
	TAG_NODE_IDLE_BEGIN,
	TAG_NODE_IDLE_END,
	TAG_NODE_LOW_BEGIN,
	TAG_NODE_LOW_END,
	TAG_NODE_HIGH_BEGIN,
	TAG_NODE_HIGH_END,
	TAG_NODE_FULL_BEGIN,
	TAG_NODE_FULL_END,
	TAG_STATUS_HEADER_BEGIN,
	TAG_STATUS_HEADER_END,
	TAG_STATUS_ROW_BEGIN,
	TAG_STATUS_ROW_END,
	TAG_STATUS_QUEUE_BEGIN,
	TAG_STATUS_QUEUE_END,
	TAG_STATUS_EXEC_BEGIN,
	TAG_STATUS_EXEC_END,
	TAG_STATUS_DIR_BEGIN,
	TAG_STATUS_DIR_END,
	TAG_STATUS_DONE_BEGIN,
	TAG_STATUS_DONE_END,
	TAG_STATUS_RUNNING_BEGIN,
	TAG_STATUS_RUNNING_END,
	TAG_STATUS_PENDING_BEGIN,
	TAG_STATUS_PENDING_END,
};

static char *tag(enum tag_t type)
{
	if (conf.color == BEN_COLOR_HTML) {
		switch (type) {
		case TAG_JOB_HEADER_BEGIN:
			return("<span class='ben-job-header'>");
		case TAG_JOB_HEADER_END:
			return("</span>\n");
		case TAG_JOB_ROW_BEGIN:
			return("");
		case TAG_JOB_ROW_END:
			return("\n");
		case TAG_JOB_EXEC_BEGIN:
			return("<span class='ben-job-exec'>");
		case TAG_JOB_REGULAR_BEGIN:
			return("<span class='ben-job-regular'>");
		case TAG_JOB_DONE_BEGIN:
			return("<span class='ben-job-done'>");
		case TAG_JOB_RUNNING_BEGIN:
			return("<span class='ben-job-running'>");
		case TAG_JOB_PENDING_BEGIN:
			return("<span class='ben-job-pending'>");
		case TAG_NODE_HEADER_BEGIN:
			return("<span class='ben-node-header'>");
		case TAG_NODE_HEADER_END:
			return("</span>\n");
		case TAG_NODE_ROW_BEGIN:
			return("");
		case TAG_NODE_ROW_END:
			return("\n");
		case TAG_NODE_CONTROL_BEGIN:
			return("<span class='ben-node-control'>");
		case TAG_NODE_COMPUTE_BEGIN:
			return("<span class='ben-node-compute'>");
		case TAG_NODE_IDLE_BEGIN:
			return("<span class='ben-node-idle'>");
		case TAG_NODE_LOW_BEGIN:
			return("<span class='ben-node-low'>");
		case TAG_NODE_HIGH_BEGIN:
			return("<span class='ben-node-high'>");
		case TAG_NODE_FULL_BEGIN:
			return("<span class='ben-node-full'>");
		case TAG_STATUS_HEADER_BEGIN:
			return("<span class='ben-status-header'>");
		case TAG_STATUS_HEADER_END:
			return("</span>\n");
		case TAG_STATUS_ROW_BEGIN:
			return("");
		case TAG_STATUS_ROW_END:
			return("\n");
		case TAG_STATUS_QUEUE_BEGIN:
			return("<span class='ben-status-queue'>");
		case TAG_STATUS_EXEC_BEGIN:
			return("<span class='ben-status-exec'>");
		case TAG_STATUS_DIR_BEGIN:
			return("<span class='ben-status-dir'>");
		case TAG_STATUS_DONE_BEGIN:
			return("<span class='ben-status-done'>");
		case TAG_STATUS_RUNNING_BEGIN:
			return("<span class='ben-status-running'>");
		case TAG_STATUS_PENDING_BEGIN:
			return("<span class='ben-status-pending'>");
		case TAG_JOB_EXEC_END:
		case TAG_JOB_REGULAR_END:
		case TAG_JOB_DONE_END:
		case TAG_JOB_RUNNING_END:
		case TAG_JOB_PENDING_END:
		case TAG_NODE_CONTROL_END:
		case TAG_NODE_COMPUTE_END:
		case TAG_NODE_IDLE_END:
		case TAG_NODE_LOW_END:
		case TAG_NODE_HIGH_END:
		case TAG_NODE_FULL_END:
		case TAG_STATUS_QUEUE_END:
		case TAG_STATUS_EXEC_END:
		case TAG_STATUS_DIR_END:
		case TAG_STATUS_DONE_END:
		case TAG_STATUS_RUNNING_END:
		case TAG_STATUS_PENDING_END:
			return("</span>");
		}
	}
	
	if (conf.color == BEN_COLOR_TTY) {
		switch (type) {
		case TAG_JOB_HEADER_BEGIN:
			return("\033[1m");
		case TAG_JOB_HEADER_END:
			return("\033[0m\n");
		case TAG_JOB_ROW_BEGIN:
			return("");
		case TAG_JOB_ROW_END:
			return("\n");
		case TAG_JOB_EXEC_BEGIN:
			return("\033[36m");
		case TAG_JOB_REGULAR_BEGIN:
		case TAG_JOB_REGULAR_END:
			return("");
		case TAG_JOB_DONE_BEGIN:
			return("\033[34m");
		case TAG_JOB_RUNNING_BEGIN:
			return("\033[1m\033[32m");
		case TAG_JOB_PENDING_BEGIN:
		case TAG_JOB_PENDING_END:
			return("");
		case TAG_NODE_HEADER_BEGIN:
			return("\033[1m");
		case TAG_NODE_HEADER_END:
			return("\033[0m\n");
		case TAG_NODE_ROW_BEGIN:
			return("");
		case TAG_NODE_ROW_END:
			return("\n");
		case TAG_NODE_CONTROL_BEGIN:
			return("\033[2m\033[34m");
		case TAG_NODE_COMPUTE_BEGIN:
		case TAG_NODE_COMPUTE_END:
			return("");
		case TAG_NODE_IDLE_BEGIN:
			return("\033[32m");
		case TAG_NODE_LOW_BEGIN:
			return("\033[36m");
		case TAG_NODE_HIGH_BEGIN:
			return("\033[33m");
		case TAG_NODE_FULL_BEGIN:
			return("\033[31m");
		case TAG_STATUS_HEADER_BEGIN:
			return("\033[1m");
		case TAG_STATUS_HEADER_END:
			return("\033[0m\n");
		case TAG_STATUS_ROW_BEGIN:
			return("");
		case TAG_STATUS_ROW_END:
			return("\n");
		case TAG_STATUS_QUEUE_BEGIN:
			return("\033[1m");
		case TAG_STATUS_EXEC_BEGIN:
			return("\033[2m");
		case TAG_STATUS_DIR_BEGIN:
			return("");
		case TAG_STATUS_DONE_BEGIN:
			return("\033[2m\033[34m");
		case TAG_STATUS_RUNNING_BEGIN:
			return("\033[1m\033[32m");
		case TAG_STATUS_PENDING_BEGIN:
			return("");
		case TAG_JOB_EXEC_END:
		case TAG_JOB_DONE_END:
		case TAG_JOB_RUNNING_END:
		case TAG_NODE_CONTROL_END:
		case TAG_NODE_IDLE_END:
		case TAG_NODE_LOW_END:
		case TAG_NODE_HIGH_END:
		case TAG_NODE_FULL_END:
		case TAG_STATUS_QUEUE_END:
		case TAG_STATUS_EXEC_END:
		case TAG_STATUS_DIR_END:
		case TAG_STATUS_DONE_END:
		case TAG_STATUS_RUNNING_END:
		case TAG_STATUS_PENDING_END:
			return("\033[0m");
		}
	}
	
	switch (type) {
	case TAG_JOB_HEADER_END:
	case TAG_JOB_ROW_END:
	case TAG_NODE_HEADER_END:
	case TAG_NODE_ROW_END:
	case TAG_STATUS_HEADER_END:
	case TAG_STATUS_ROW_END:
		return("\n");
	default:
		return("");
	}
}

static void control_header_job(int flags)
{
	printf("%s", tag(TAG_JOB_HEADER_BEGIN));
	
	printf("#%5s %17s %25s %1s %6s %-9s",
		"id", "dir", "job", "S", "node", "");

	if (flags & BEN_FLAG_DURATIONS) {
		printf("  %8s", "duration");
	}
	
	if (flags & BEN_FLAG_TIMES) {
		printf("   %19s   %19s", "start", "stop");
	}

	printf("%s", tag(TAG_JOB_HEADER_END));
}

static int control_print_job(control *k, int flags)
{
	int tid, brestr, bid, type;
	char *tdir, *tname, *bname;
	uint64_t start, stop;
	char bname_short[10];
	gtime_str ddur, dstart, dstop;
	
	if (proto_fetch(&k->inbuff, "issiiispp", &tid, &tdir, &tname, &type,
			&brestr, &bid, &bname, &start, &stop)) {
		DEBUG("%s: PROTO_SV_JOB: bad packet\n", k->action);
		return(-1);
	}
	
	if (flags & BEN_FLAG_GATHER) {
		control_gather_job(k, tdir, type, (brestr >= 0), start, stop);
		return(0);
	}
	
	printf("%s", tag(TAG_JOB_ROW_BEGIN));
	
	printf("%6d %17s %s%25s%s ", tid, tdir,
		tag((brestr >= 0) ? TAG_JOB_EXEC_BEGIN : TAG_JOB_REGULAR_BEGIN),
		tname,
		tag((brestr >= 0) ? TAG_JOB_EXEC_END : TAG_JOB_REGULAR_END));
	
	if (type == PROTO_TASK_DONE)
		printf("%s.%s",
			tag(TAG_JOB_DONE_BEGIN), tag(TAG_JOB_DONE_END));
	else if (type == PROTO_TASK_RUNNING)
		printf("%sr%s",
			tag(TAG_JOB_RUNNING_BEGIN), tag(TAG_JOB_RUNNING_END));
	else
		printf("%s %s",
			tag(TAG_JOB_PENDING_BEGIN), tag(TAG_JOB_PENDING_END));
	
	size_t l = strlen(bname);
	if (l <= 9) {
		memcpy(bname_short, bname, l + 1);
	} else {
		memcpy(bname_short, bname, 8);
		bname_short[8] = '+';
		bname_short[9] = 0;
	}
	
	if (bid == -1)
		printf(" %6s %-9s", "", bname_short);
	else
		printf(" %6d %-9s", bid, bname_short);
	
	if (flags & BEN_FLAG_DURATIONS) {
		if (start)
			ddur = duration_str(gtime_s(stop - start), 0);
		else
			ddur.s[0] = 0;
		
		printf(" %s%8s",
			(type == PROTO_TASK_RUNNING) ? ">" : " ",
			ddur.s);
	}
	
	if (flags & BEN_FLAG_TIMES) {
		if (start)
			dstart = date_str(gtime_s(start), 0);
		else
			dstart.s[0] = 0;

		if (stop)
			dstop = date_str(gtime_s(stop), 0);
		else
			dstop.s[0] = 0;
		
		printf("   %19s   %19s", dstart.s, dstop.s);
	}
	
	printf("%s", tag(TAG_JOB_ROW_END));
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
static void control_header_node(int flags)
{
	printf("%s", tag(TAG_NODE_HEADER_BEGIN));
	
	printf("#%5s %-50s  R  P", "node", "");
	
	if (flags & BEN_FLAG_VERBOSE) {
		printf("   in  out ref flags");
	}
	
	printf("%s", tag(TAG_NODE_HEADER_END));
}

static int control_print_node(control *k, int flags)
{
	int bid, run, n, refs, bflags, load0, load1;
	char *bname;
	uint64_t inbsz, outbsz;
	
	if (proto_fetch(&k->inbuff, "isiippii", &bid, &bname,
			&run, &n, &inbsz, &outbsz, &refs, &bflags)) {
		DEBUG("%s: PROTO_SV_NODE: bad packet\n", k->action);
		return(-1);
	}
	
	printf("%s", tag(TAG_NODE_ROW_BEGIN));

	if ((n < 0) && (run == 0)) {
		printf("%s %5d %-50s %2s %2s%s",
			tag(TAG_NODE_CONTROL_BEGIN),
			bid, bname, "-", "C",
			tag(TAG_NODE_CONTROL_END));
	} else {
		if (run == n) {
			load0 = TAG_NODE_FULL_BEGIN;
			load1 = TAG_NODE_FULL_END;
		} else if (run == 0) {
			load0 = TAG_NODE_IDLE_BEGIN;
			load1 = TAG_NODE_IDLE_END;
		} else if (run <= n / 2) {
			load0 = TAG_NODE_LOW_BEGIN;
			load1 = TAG_NODE_LOW_END;
		} else {
			load0 = TAG_NODE_HIGH_BEGIN;
			load1 = TAG_NODE_HIGH_END;
		}
		
		printf(" %5d %-50s %s%2d%s %2d", bid, bname,
			tag(load0), run, tag(load1), n);
	}
	
	if (flags & BEN_FLAG_VERBOSE) {
		printf(" %4ld %4ld %3d %c%c%c",
			(long)inbsz, (long)outbsz, refs,
			(bflags & NODE_KICK) ? 'k' : '-',
			(bflags & NODE_SYNC) ? 'S' : '-',
			(bflags & NODE_RETIRE) ? 'R' : '-');
	}
	
	printf("%s", tag(TAG_NODE_ROW_END));
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
static void control_print_odir(char *name, uint64_t start, uint64_t stop,
	int is_exec, int done, int running, int pending, int flags)
{
	gtime_str ddur, dstart, dstop;
	int d0, d1;
	int total;
	
	if (start) {
		dstart = date_str(gtime_s(start), 0);
		dstop = date_str(gtime_s(stop), 0);
		ddur = duration_str(gtime_s(stop - start), GTIME_DAYS);
	} else {
		dstart.s[0] = 0;
		dstop.s[0] = 0;
		ddur.s[0] = 0;
	}
	
	total = done + running + pending;
	
	if (is_exec) {
		d0 = TAG_STATUS_EXEC_BEGIN;
		d1 = TAG_STATUS_EXEC_END;
		name = "[exec]";
	} else if (name == NULL) {
		d0 = TAG_STATUS_QUEUE_BEGIN;
		d1 = TAG_STATUS_QUEUE_END;
		name = "[queue]";
		
	} else {
		d0 = TAG_STATUS_DIR_BEGIN;
		d1 = TAG_STATUS_DIR_END;
	}
	
	printf("%s%s%17s%s",
		tag(TAG_STATUS_ROW_BEGIN), tag(d0), name, tag(d1));

	printf(" %s", tag(TAG_STATUS_DONE_BEGIN));

	if ((done) || (running)) {
		printf("%8d", done);
		if (done < total)
			printf(" %2d%%", 100 * done / total);
		else
			printf("    ");
	} else {
		printf("%8s %2s ", "", "");
	}
	
	printf("%s %s",
		tag(TAG_STATUS_DONE_END), tag(TAG_STATUS_RUNNING_BEGIN));

	if (running)
		printf("%8d", running);
	else
		printf("%8s", "");
	
	printf("%s %s",
		tag(TAG_STATUS_RUNNING_END), tag(TAG_STATUS_PENDING_BEGIN));

	if (pending)
		printf("%8d", pending);
	else
		printf("%8s", "");

	printf("%s",
		tag(TAG_STATUS_PENDING_END));

	if (flags & BEN_FLAG_DURATIONS) {
		printf("   %s%14s",
			((running) || (start && pending)) ? ">" : " ",
			ddur.s);
	}

	if (flags & BEN_FLAG_TIMES) {
		printf("   %19s   %19s",
			dstart.s, dstop.s);
	}
	
	printf("%s", tag(TAG_STATUS_ROW_END));
}

static void control_print_status(control *k, int flags)
{
	odir_info *odir;
	int n, o;
	
	printf("%s", tag(TAG_STATUS_HEADER_BEGIN));
	
	printf("#%16s %8s %2s  %8s %8s",
		"dir", "done", "", "running", "pending");

	if (flags & BEN_FLAG_DURATIONS) {
		printf("    %14s", "duration");
	}
	
	if (flags & BEN_FLAG_TIMES) {
		printf("   %19s   %19s", "first-start", "last-stop");
	}

	printf("%s", tag(TAG_STATUS_HEADER_END));

	control_print_odir(NULL, k->g_start, k->g_stop,
		0, k->g_done, k->g_running, k->g_pending, flags);
	
	odir = (odir_info *)k->odir.data;
	n = k->nodir;
	
	for (o = 0; o < n; o++) {
		control_print_odir(odir[o].name, odir[o].start, odir[o].stop,
			odir[o].is_exec,
			odir[o].done, odir[o].running, odir[o].pending,
			flags);
	}
}

// **************************************************************************
// 
// **************************************************************************
static int control_iter(control *k, int flags)
{
	void *d;
	size_t dsz;
	uint64_t ping_id;
	int tid, cid;
	char *hs, *s;
	int h;
	
	if (proto_prepare(&k->inbuff, PROTO_SV_MAGIC, conf.proto_buff_max)) {
		DEBUG("%s: protocol error\n", k->action);
		return(-1);
	}
	
	if (!proto_ready(&k->inbuff))
		return(0);
	
	h = proto_header(&k->inbuff);

	// ************************************	
	if (h == PROTO_SV_MSG) {
		if (proto_fetch(&k->inbuff, "d", &d, &dsz)) {
			DEBUG("%s: PROTO_SV_MSG: bad packet\n", k->action);
			return(-1);
		}

		if ((dsz != 0) && (fwrite(d, dsz, 1, stdout) != 1)) {
			DEBUG("fwrite(%zd): %s\n", dsz, strerror(errno));
			return(-1);
		}
		
		return(0);
	}

	// ************************************	
	if ((h == PROTO_SV_OUT) || (h == PROTO_SV_ERR)) {
		hs = (h == PROTO_SV_OUT) ? "OUT" : "ERR";
		
		if (proto_fetch(&k->inbuff, "iisd", &tid, &cid, &s, &d, &dsz)) {
			DEBUG("%s: PROTO_SV_%s: bad packet\n", k->action, hs);
			return(-1);
		}
		
		control_print(h, tid, cid, s, (char *)d, dsz);
		
		return(0);
	}

	// ************************************	
	if (h == PROTO_SV_PING) {
		if (proto_fetch(&k->inbuff, "p", &ping_id)) {
			DEBUG("%s: PROTO_SV_PING: bad packet\n", k->action);
			return(-1);
		}

		proto_append(&k->outbuff, PROTO_CL_MAGIC, PROTO_CL_PONG,
			"p", ping_id);
		
		//DEBUG("[ping] %zd\n", ping_id);
		return(0);
	}

	// ************************************	
	if (h == PROTO_SV_JOB) {
		return(control_print_job(k, flags));
	}

	// ************************************	
	if (h == PROTO_SV_NODE) {
		return(control_print_node(k, flags));
	}

	// ************************************	
	DEBUG("%s: unknown response header 0x%08x\n",
		k->action, h);
	return(-1);
}

// **************************************************************************
// 
// **************************************************************************
static int control_loop(control *k, int flags)
{
	struct pollfd pfd;
	int done, error, detached;
	
	done = 0;
	error = 0;
	detached = 0;
	
	while (1) {
		pfd.fd = k->fd;
		pfd.events = POLLIN;
		if (buffer_io_complete(&k->outbuff)) {
			if ((!detached) && (flags & BEN_FLAG_DAEMONIZE)) {
				if (exec_daemonize(NULL, NULL, NULL))
					return(1);
				detached = 1;
			}
		} else {
			pfd.events |= POLLOUT;
		}
		pfd.revents = 0;
	
		if (poll(&pfd, 1, -1) < 0) {
			fprintf(stderr, "%s: poll(): %s\n",
				k->action, strerror(errno));
			return(1);
		}
		
		if (pfd.revents & (POLLIN|POLLERR|POLLHUP)) {
			buffer_io_read(&k->inbuff, k->fd);
			done |= k->inbuff.eof;
			error |= k->inbuff.error;
		}
		
		if (pfd.revents & (POLLOUT|POLLERR|POLLHUP)) {
			buffer_io_write(&k->outbuff, k->fd);
			error |= k->outbuff.eof | k->outbuff.error;
		}
		
		if ((done) || (error)) {
			error |= (k->inbuff.partial != 0);
			error |= !buffer_io_complete(&k->outbuff);

			if (error) {
				fprintf(stderr, "%s: i/o error\n", k->action);
				return(1);
			} else {
				// fprintf(stderr, "-- ok --\n");
				return(0);
			}
		}
		
		if (control_iter(k, flags))
			return(1);
	}
}

// **************************************************************************
// 
// **************************************************************************
int ben_add(char *addr, char *dir, char *cmd, char *opts,
	int n, char **names)
{
	control k;
	int i, r;
	
	if (control_init(&k, "ben add", addr))
		return(1);
	
	if (!dir)
		dir = "";
	
	if (!cmd)
		cmd = "";
	
	if (!opts)
		opts = "";
	
	for (i = 0; i < n; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_ADD,
			"sssss", conf.shell, dir, opts, names[i], cmd);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");
	
	r = control_loop(&k, 0);
	
	control_clear(&k);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int ben_rm(char *addr, int mask, char *dir, int n, char **names)
{
	control k;
	char *defz;
	int i, r;
	
	if (control_init(&k, "ben stop", addr))
		return(1);
	
	if (!dir)
		dir = "";

	if (n == 0) {
		defz = "";
		names = &defz;
		n = 1;
	}
	
	for (i = 0; i < n; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_RM,
			"iss", mask, dir, names[i]);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");

	r = control_loop(&k, 0);
	
	control_clear(&k);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int ben_exec(char *addr, char *cmd, char *opts, int flags)
{
	control k;
	int r;
	
	if (control_init(&k, "ben exec", addr))
		return(1);
	
	if (!cmd)
		cmd = "";
	
	if (!opts)
		opts = "";
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_EXEC,
		"sss", conf.shell, opts, cmd);
	
	r = control_loop(&k, flags);
	
	control_clear(&k);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
#define JOB_ENTRIES_SKIP	\
		((int)(sizeof(job_entry_skip) / sizeof(job_entry_skip[0])))

static const char *job_entry_skip[] = {
	"id", "running_id", "running_name", "restrict_id", "forward_id",
	"forward_name", "ran_id", "ran_name", "stdout_path", "stderr_path",
	"start_time", "stop_time"
};

typedef struct jobspec_t {
	char *type;
	char *shell;
	char *dir;
	char *options;
	char *name;
	char *command;
	char *restrict_name;
} jobspec;

jobspec job_def(char *dir, char *cmd, char *opts)
{
	jobspec j;
	
	j.type = NULL;
	j.shell = conf.shell;
	j.dir = dir;
	j.command = cmd;
	j.options = opts;
	
	j.name = NULL;
	
	j.restrict_name = NULL;
	
	return j;
}

void job_print(jobspec j)
{
	fprintf(stderr, " %20s: %s", "type",
		(j.type) ? j.type : "<default>");
	fprintf(stderr, " %20s: %s", "shell", (j.shell) ? j.shell : "N/A");
	fprintf(stderr, " %20s: %s", "dir", (j.dir) ? j.dir : "N/A");
	fprintf(stderr, " %20s: %s", "options",
		(j.options) ? j.options : "N/A");
	fprintf(stderr, " %20s: %s", "name", (j.name) ? j.name : "N/A");
	fprintf(stderr, " %20s: %s", "command",
		(j.command) ? j.command : "N/A");
	fprintf(stderr, " %20s: %s", "node",
		(j.restrict_name) ? j.restrict_name : "<any>");
}

int job_submit(control *k, jobspec j)
{
	// all fields must be non-nil except:
	// - type
	// - restrict_name
	if ((!j.shell) || (!j.dir) || (!j.options) || (!j.name)
	||  (!j.command)) {
		fprintf(stderr, "Incomplete job specification\n");
		job_print(j);
		return -1;
	}
	
	if ((j.type) && (!strcmp(j.type, "done"))) {
		// jobs that have type = "done" are skipped
		return 0;
	}
	
	proto_append(&k->outbuff, PROTO_CL_MAGIC, PROTO_CL_RESUME,
		"ssssss", j.shell, j.dir, j.options, j.name, j.command,
		(j.restrict_name) ? j.restrict_name : "");

	return 0;
}

static int ben_resume_file(control *k, buffer *b,
	char *dir, char *cmd, char *opts, char *path)
{
	if (buffer_load_path(b, path, conf.file_size_max))
		return -1;
	
	jobspec j = job_def(dir, cmd, opts);
	int mode = 0;
	
	while (1) {
		char *entry;
		char *value;
		
		if (ini_get(b, &entry, &value))
			return -1;
		
		// eof
		if (!entry) {
			if (mode == 1) {
				if (job_submit(k, j))
					return -1;
			}
			break;
		}
		
		// section
		if (!value) {
			if (mode == 1) {
				if (job_submit(k, j))
					return -1;
			}
			
			if (strcmp(entry, "job") != 0) {
				fprintf(stderr, "%s: Warning: unknown "
					"section '%s'\n", path, entry);
				mode = 2;
				continue;
			}
			
			mode = 1;
			j = job_def(dir, cmd, opts);
			continue;
		}
		
		// entry
		if (mode != 1)
			continue;
		
		if (!strcmp(entry, "type")) {
			j.type = value;
			continue;
		}
		
		if (!strcmp(entry, "shell")) {
			j.shell = value;
			continue;
		}
		
		if (!strcmp(entry, "dir")) {
			j.dir = value;
			continue;
		}
		
		if (!strcmp(entry, "options")) {
			j.options = value;
			continue;
		}
		
		if (!strcmp(entry, "name")) {
			j.name = value;
			continue;
		}
		
		if (!strcmp(entry, "command")) {
			j.command = value;
			continue;
		}

		if (!strcmp(entry, "restrict_name")) {
			j.restrict_name = value;
			continue;
		}
		
		int ignore = 0;
		for (int e = 0; e < JOB_ENTRIES_SKIP; e++) {
			if (!strcmp(entry, job_entry_skip[e])) {
				ignore = 1;
				break;
			}
		}
		
		if (ignore)
			continue;
		
		fprintf(stderr, "%s: Warning: unknown entry '%s' "
			"(value '%s')\n",
			path, entry, value);
	}
	
	return 0;
}

int ben_resume(char *addr, char *dir, char *cmd, char *opts,
	int n, char **paths, int flags)
{
	control k;
	
	if (control_init(&k, "ben resume", addr))
		return(1);
	
	for (int i = 0; i < n; i++) {
		buffer b;
		
		buffer_init(&b, 0);
		
		int r = ben_resume_file(&k, &b, dir, cmd, opts, paths[i]);
		
		buffer_clear(&b);
		
		if (r)
			return -1;
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");
	
	int r = control_loop(&k, flags);
	
	control_clear(&k);
	
	return(r);
}


// **************************************************************************
// 
// **************************************************************************
int ben_scale(char *addr, int n, int nb, char **b, int flags)
{
	control k;
	int i, r;
	
	if (control_init(&k, "ben scale", addr))
		return(1);
	
	flags &= PROTO_SCALE_KILL | PROTO_SCALE_RETIRE;
	
	if (nb == 0) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_SCALE,
			"sii", "", n, flags);
	}
	
	for (i = 0; i < nb; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_SCALE,
			"sii", b[i], n, flags);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");

	r = control_loop(&k, 0);
	
	control_clear(&k);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int ben_list(char *addr, int flags, char *dir, int n, char **names)
{
	control k;
	char *defz;
	exec_pager pager;
	int i, r;
	
	if (control_init(&k, "ben list", addr))
		return(1);
	
	if (!dir)
		dir = "";
	
	if (n == 0) {
		defz = "";
		names = &defz;
		n = 1;
	}
	
	for (i = 0; i < n; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_LIST,
			"iss", flags, dir, names[i]);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");

	exec_pager_run(&pager);
	
	if (!(flags & BEN_FLAG_GATHER))
		control_header_job(flags);
	
	r = control_loop(&k, flags);
	
	if (flags & BEN_FLAG_GATHER)
		control_print_status(&k, flags);
	
	exec_pager_wait(&pager);
	
	control_clear(&k);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int ben_nodes(char *addr, int flags)
{
	control k;
	exec_pager pager;
	int r;
	
	if (control_init(&k, "ben nodes", addr))
		return(1);
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_NODES, "");
	
	exec_pager_run(&pager);
	
	control_header_node(flags);

	r = control_loop(&k, flags);

	exec_pager_wait(&pager);

	control_clear(&k);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int ben_exit(char *addr, int flags)
{
	control k;
	int r;
	
	if (control_init(&k, "ben exit", addr))
		return(1);
	
	flags &= PROTO_SCALE_RETIRE;
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_EXIT, "i", flags);
	
	r = control_loop(&k, 0);
	
	control_clear(&k);
	
	return(r);
}
