/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "config.h"
#include "xmem.h"
#include "strlist.h"
#include "server.h"
#include "client.h"
#include "control.h"
#include "arg.h"
#include "common.h"
#include "exec.h"
#include "man.h"
#include "ssh.h"
#include "main.h"

// ****************************************************************************
// Help
// ****************************************************************************
//#define HELP_STRINGIFY2(s)	#s
//#define HELP_STRINGIFY(s)	HELP_STRINGIFY2(s)

#define HELP_GENERAL	\
"ben {command} [-s path|host:port] [-x ext] [-f user@host] " \
     "[--chdir dir] [-d]\n" \

#define HELP_SERVER	\
"ben server [--snapshot path]\n"

#define HELP_CLIENT	\
"ben client [--name name] [-n processes] [-r user@host]\n"

#define HELP_ADD	\
"ben add [-o dir] {-c command|-C file} [-q options|-Q file] " \
"{name...|-J file|-i file}\n"

#define HELP_RM		\
"ben rm [-o dir] [-t d|r|p...] [name...|-J file]\n"

#define HELP_EXEC	\
"ben exec {-c command|-C file} [-q options|-Q file]\n"

#define HELP_SCALE	\
"ben scale|kill [node...|-B file] [-n processes] [--retire]\n"

#define HELP_LIST	\
"ben list|status [-l] [-o dir] [-t d|r|p...] [name...|-J file]\n"

#define HELP_NODES	\
"ben nodes [-v]\n"

#define HELP_EXIT	\
"ben exit [--retire]\n"

#define HELP_ALL	\
"ben v" BEN_VERSION " -- Laurent Poirrier, " BEN_YEARS "\n" \
"\n" \
"General options:\n" \
"\n" \
"  " HELP_GENERAL \
"\n" \
"Commands:\n" \
"\n" \
"  " HELP_SERVER \
"\n" \
"  " HELP_CLIENT \
"\n" \
"  " HELP_ADD \
"  " HELP_RM \
"  " HELP_EXEC \
"  " HELP_SCALE \
"  " HELP_LIST \
"  " HELP_NODES \
"  " HELP_EXIT \
"\n"


// ****************************************************************************
// Job statuses
// ****************************************************************************
static int parse_status(char *str)
{
	char *s;
	int mask;
	
	if (!str) {
		mask = PROTO_TASK_DONE
			| PROTO_TASK_RUNNING | PROTO_TASK_PENDING;
		return(mask);
	}
	
	mask = 0;
	s = str;
	
	while (*s != 0) {
		if (*s == 'd') {
			mask |= PROTO_TASK_DONE;
		} else if (*s == 'r') {
			mask |= PROTO_TASK_RUNNING;
		} else if (*s == 'p') {
			mask |= PROTO_TASK_PENDING;
		} else {
			fprintf(stderr,
				"%s: Unknown job status '%c'.\n",
				str, *s);
			return(-1);
		}
		s++;
	}
	
	return(mask);
}

// ****************************************************************************
// User@host
// ****************************************************************************
static char *parse_user(args *a, char *dst)
{
	char *at, *user;
	
	at = strchr(dst, '@');
	
	if (at) {
		user = xmalloc(at - dst + 1);
		memcpy(user, dst, at - dst);
		user[at - dst] = 0;
		arg_hook(a, user);
		return(user);
	}
	
	return(NULL);
}

// ****************************************************************************
// Addresses
// ****************************************************************************
#define BEN_ADDR_MIDPOINT	1
static int parse_address(args *a, char **addr, char *dst,
	char *ss, char *ls, char *sx, char *lx, int flags)
{
	socket_addr r;
	char *d, *user, *e;
	char *s, *x;
	int port;
	
	// parse user
	if (dst) {
		user = parse_user(a, dst);
		
		if (!user)
			user = conf.user;
	} else {
		user = conf.user;
	}
	
	// full address
	s = arg_str(a, ss, ls, NULL, NULL, NULL);
	
	if (s) {
		if (socket_addr_parse(&r, s)) {
			fprintf(stderr, "%s: bad address\n", s);
			return(-1);
		}
		
		// Unix
		if (r.type == SOCKET_UNIX) {
			if (addr)
				*addr = s;
			return(0);
		}
		
		// TCP/IP
		if (r.serv[0] != 0) {
			if (addr)
				*addr = s;
			return(0);
		}
		
		d = xprintf(NULL, "%s:%d",
			r.path, conf.default_port);
		arg_hook(a, d);
		if (addr)
			*addr = d;
		return(0);
	}
	
	// extension
	x = arg_str(a, sx, lx, NULL, NULL, NULL);
	
	if (x) {
		// Unix
		if (x[0] != ':') {
			d = xprintf(NULL, "/tmp/ben-%s/socket-%s",
				user, x);
			arg_hook(a, d);
			if (addr)
				*addr = d;
			return(0);
		}
		
		// TCP/IP
		port = strtol(x + 1, &e, 0);
		
		if (*e != 0) {
			fprintf(stderr, "%s: bad port number\n",
				x + 1);
			return(-1);
		}
		
		d = xprintf(NULL, "localhost:%d",
			conf.default_port + port);
		arg_hook(a, d);
		if (addr)
			*addr = d;
		return(0);
	}
	
	// midpoint
	if (flags & BEN_ADDR_MIDPOINT) {
		d = xprintf(NULL, "/tmp/ben-%s/socket-%s",
			user, conf.random_id);
		arg_hook(a, d);
		if (addr)
			*addr = d;
		return(0);
	}
	
	// default
	d = xprintf(NULL, "/tmp/ben-%s/socket-default", user);
	arg_hook(a, d);
	if (addr)
		*addr = d;
	return(0);
}

// ****************************************************************************
// Global options
// ****************************************************************************
typedef struct gopts_t {
	char *cmd;
	char *addr;
	char *fdst, *faddr;
	char *cdir;
	int flags;
} gopts;

// ****************************************************************************
// Read global options
// ****************************************************************************
static int main_global(args *a, gopts *opt)
{
	char *s;
	
	// init
	memset(opt, 0, sizeof(gopts));
	
	// chdir
	opt->cdir = arg_str(a, NULL, "chdir", NULL, NULL, NULL);

	// flags
	opt->flags = 0;

	if (arg_switch(a, "d", "daemon"))
		opt->flags |= BEN_FLAG_DAEMONIZE;
	
	if (arg_switch(a, NULL, "bridge"))
		opt->flags |= BEN_FLAG_BRIDGE;
	
	// color
	s = arg_str(a, NULL, "color", NULL, NULL, NULL);
	
	if ((!s) || (!strcmp(s, "auto"))) {
		// keep conf.color as it is
	} else if ((!strcmp(s, "always")) || (!strcmp(s, "yes"))
	|| (!strcmp(s, "1"))) {
		conf.color = BEN_COLOR_TTY;
	} else if ((!strcmp(s, "never")) || (!strcmp(s, "no"))
	|| (!strcmp(s, "0"))) {
		conf.color = BEN_COLOR_NONE;
	} else if (!strcmp(s, "html")) {
		conf.color = BEN_COLOR_HTML;
	}
	
	// --socket and --forward addresses
	opt->fdst = arg_str(a, "f", "forward", NULL, NULL, NULL);
	
	if (opt->fdst) {
		if (parse_address(a, &opt->faddr, opt->fdst,
				"s", "socket", "x", "extension", 0))
			return(-1);

		if (parse_address(a, &opt->addr, NULL,
				NULL, "local-socket", NULL, "local-extension",
				(opt->flags & BEN_FLAG_BRIDGE)
					? 0 : BEN_ADDR_MIDPOINT))
			return(-1);
	} else {
		if (parse_address(a, &opt->addr, NULL,
				"s", "socket", "x", "extension", 0))
			return(-1);

		opt->faddr = NULL;
	}
	
	// command
	opt->cmd = arg_str(a, "", "", NULL, NULL, NULL);
	
	return(0);
}

// ****************************************************************************
// Server
// ****************************************************************************
static int main_server(args *a, gopts *opt)
{
	char *snapshot_path = arg_str(a, NULL, "snapshot", NULL, NULL, NULL);
	
	if (arg_finish(a))
		return(-1);
	
	if (opt->cdir) {
		if (exec_mkchdir(opt->cdir))
			return(1);
	}
	
	return(ben_server(opt->addr, snapshot_path, opt->flags));
}

// ****************************************************************************
// Remote client
// ****************************************************************************
static int main_remote_client(args *a, gopts *opt, char *rdst)
{
	exec_unit fu;
	char *raddr, *rbin;
	char *name;
	char *cmd, *resp;
	int n, r;
	
	// client options
	name = arg_str(a, NULL, "name", NULL, NULL, NULL);
	
	n = arg_int(a, "n", "processes", 1);
	
	// remote options
	rbin = arg_str(a, NULL, "remote-path", NULL, NULL, "ben");
	
	if (parse_address(a, &raddr, rdst,
			NULL, "remote-socket", NULL, "remote-extension",
			(opt->flags & BEN_FLAG_BRIDGE)
				? 0 : BEN_ADDR_MIDPOINT))
		return(-1);
	
	// finish
	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}
	
	// remote
	cmd = xprintf(NULL, "%s client --socket %s %s%s %s%s -n %d %s",
		rbin, raddr,
		(opt->cdir) ? "--chdir " : "",
		(opt->cdir) ? opt->cdir : "",
		(name) ? "--name " : "",
		(name) ? name : "",
		n,
		(opt->flags & BEN_FLAG_BRIDGE) ? "--bridge" : "");
	
	resp = (opt->flags & BEN_FLAG_DAEMONIZE) ? BEN_CLIENT_ACK_STR : NULL;
	
	r = ssh_remote_exec(rdst, raddr, opt->addr, cmd, resp,
		conf.remote_log);

	// cleanup
	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);
	
	return(r);
}

// ****************************************************************************
// Client
// ****************************************************************************
static int main_client(args *a, gopts *opt)
{
	exec_unit fu;
	char *rdst;
	char *name;
	int n, r;
	
	// remote
	rdst = arg_str(a, "r", "remote", NULL, NULL, NULL);
	
	if (rdst)
		return(main_remote_client(a, opt, rdst));

	// client options
	name = arg_str(a, NULL, "name", NULL, NULL, conf.host);
	
	n = arg_int(a, "n", "processes", INT_MIN);
	
	if (n == INT_MIN) {
		if (opt->flags & BEN_FLAG_BRIDGE)
			n = -1;
		else
			n = 1;
	} else if (n < 0) {
		n = conf.cpus_online;
	}
	
	// finish
	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}
	
	if (opt->cdir) {
		if (exec_mkchdir(opt->cdir))
			return(1);
	}

	r = ben_client(opt->addr, name, n, opt->flags);
	
	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);
	
	return(r);
}

// ****************************************************************************
// Add
// ****************************************************************************
static int main_add(args *a, gopts *opt)
{
	exec_unit fu;
	strlist names;
	strlist inputs;
	char *odir;
	char *command;
	char *options;
	int r = 0;
	
	odir = arg_str(a, "o", "outdir", NULL, NULL, NULL);
	command = arg_str(a, "c", "command", "C", "command-file", NULL);
	options = arg_str(a, "q", "options", "Q", "options-file", NULL);
	
	strlist_init(&inputs);
	strlist_init(&names);
	
	arg_strlist(a, "i", "input", NULL, NULL, &inputs);
	arg_strlist(a, "j", "job", "J", "job-file", &names);
	arg_strlist(a, "", "", NULL, NULL, &names);
	
	if ((names.n < 1) && (inputs.n < 1)) {
		printf(HELP_ADD);
		return(0);
	}
	
	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	
	if (names.n >= 1) {
		r |= ben_add(opt->addr, odir, command, options,
			names.n, strlist_data(&names));
	}
	
	if (inputs.n >= 1) {
		r |= ben_resume(opt->addr, odir, command, options,
			inputs.n, strlist_data(&inputs), opt->flags);
	}
	
	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	strlist_clear(&names);
	strlist_clear(&inputs);

	return(r);
}

// ****************************************************************************
// Remove
// ****************************************************************************
static int main_rm(args *a, gopts *opt)
{
	exec_unit fu;
	strlist names;
	char *odir;
	char *s;
	int r, mask;

	// 'ben rm' without argument is actually valid and would
	// remove all tasks. But as it is somewhat dangerous, we
	// display the help in that case. Use 'ben rm -t drp'
	// instead.
	if (arg_remain(a) == 0) {
		printf(HELP_RM);
		return(0);
	}
	
	odir = arg_str(a, "o", "outdir", NULL, NULL, NULL);
	
	s = arg_str(a, "t", "status", NULL, NULL, NULL);
	mask = parse_status(s);
	if (mask == -1)
		return(-1);
	
	strlist_init(&names);
	
	arg_strlist(a, "j", "job", "J", "job-file", &names);
	arg_strlist(a, "", "", NULL, NULL, &names);

	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	r = ben_rm(opt->addr, mask, odir,
		names.n, strlist_data(&names));

	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	strlist_clear(&names);

	return(r);
}

// ****************************************************************************
// Exec
// ****************************************************************************
static int main_exec(args *a, gopts *opt)
{
	exec_unit fu;
	char *command;
	char *options;
	int r;
	
	command = arg_str(a, "c", "command", "C", "command-file", NULL);
	options = arg_str(a, "q", "options", "Q", "options-file", NULL);
	
	if (!command) {
		printf(HELP_EXEC);
		return(0);
	}
	
	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	r = ben_exec(opt->addr, command, options, opt->flags);

	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	return(r);
}

// ****************************************************************************
// Exec
// ****************************************************************************
static int main_scale(args *a, gopts *opt, int kill)
{
	exec_unit fu;
	strlist names;
	int n, r;
	
	// 'ben scale' without argument is actually valid and would
	// set all nodes to 0 processes. But as it is probably not useful,
	// we display the help in that case. Use 'ben scale -n 0' instead.
	if ((!kill) && (arg_remain(a) == 0)) {
		printf(HELP_SCALE);
		return(0);
	}
	
	strlist_init(&names);
	
	if (kill)
		opt->flags |= PROTO_SCALE_KILL;
	
	n = arg_int(a, "n", "processes", 0);
	
	if (arg_switch(a, NULL, "retire"))
		opt->flags |= PROTO_SCALE_RETIRE;
	
	arg_strlist(a, "b", "node", "B", "node-file", &names);
	arg_strlist(a, "", "", NULL, NULL, &names);

	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	r = ben_scale(opt->addr, n, names.n, strlist_data(&names),
		opt->flags);

	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	return(r);
}

// ****************************************************************************
// List tasks & status
// ****************************************************************************
static int main_list(args *a, gopts *opt, int status_cmd)
{
	exec_unit fu;
	strlist names;
	char *odir;
	int mask, mask_opts, r;
	char *s;

	mask_opts = BEN_FLAG_DURATIONS;
	
	if (status_cmd)
		mask_opts |= BEN_FLAG_GATHER;
	
	if (arg_switch(a, "l", "long"))
		mask_opts |= BEN_FLAG_TIMES;
	
	odir = arg_str(a, "o", "outdir", NULL, NULL, NULL);
	
	s = arg_str(a, "t", "status", NULL, NULL, NULL);
	mask = parse_status(s);
	if (mask == -1)
		return(-1);

	strlist_init(&names);
	
	arg_strlist(a, "j", "job", "J", "job-file", &names);
	arg_strlist(a, "", "", NULL, NULL, &names);

	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	r = ben_list(opt->addr, mask | mask_opts,
		odir, names.n, strlist_data(&names));
	
	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	strlist_clear(&names);

	return(r);
}

// ****************************************************************************
// Nodes info
// ****************************************************************************
static int main_nodes(args *a, gopts *opt)
{
	exec_unit fu;
	int opts, r;
	
	opts = 0;
	
	if (arg_switch(a, "v", "verbose"))
		opts |= BEN_FLAG_VERBOSE;

	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	r = ben_nodes(opt->addr, opts);
	
	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	return(r);
}

// ****************************************************************************
// Exit server
// ****************************************************************************
static int main_exit(args *a, gopts *opt)
{
	exec_unit fu;
	int r;
	
	if (arg_switch(a, NULL, "retire"))
		opt->flags |= PROTO_SCALE_RETIRE;

	if (arg_finish(a))
		return(-1);

	if (opt->fdst) {
		if (ssh_fwd_fork(&fu, opt->fdst, opt->addr, opt->faddr))
			return(-1);
	}

	r = ben_exit(opt->addr, opt->flags);
	
	if (opt->fdst)
		r |= ssh_fwd_terminate(&fu);

	return(r);
}

// ****************************************************************************
// Command-line arguments
// ****************************************************************************
static int main_args(args *a, gopts *opt)
{
	if (main_global(a, opt))
		return(-1);

	if (!opt->cmd) {
		printf(HELP_ALL);
		return(0);
	}
	
	if (!strcmp(opt->cmd, "help")) {
		printf(HELP_ALL);
		return(0);
	}

	if (!strcmp(opt->cmd, "man")) {
		if (arg_finish(a))
			return(-1);
		return(man("ben"));
	}
	
	if (!strcmp(opt->cmd, "server"))
		return(main_server(a, opt));
	
	if (!strcmp(opt->cmd, "client"))
		return(main_client(a, opt));

	if (!strcmp(opt->cmd, "add"))
		return(main_add(a, opt));
	
	if (!strcmp(opt->cmd, "exec"))
		return(main_exec(a, opt));
	
	if (!strcmp(opt->cmd, "rm"))
		return(main_rm(a, opt));
	
	if (!strcmp(opt->cmd, "scale"))
		return(main_scale(a, opt, 0));
	
	if (!strcmp(opt->cmd, "kill"))
		return(main_scale(a, opt, 1));
	
	if (!strcmp(opt->cmd, "list"))
		return(main_list(a, opt, 0));
	
	if (!strcmp(opt->cmd, "status"))
		return(main_list(a, opt, 1));

	if (!strcmp(opt->cmd, "nodes"))
		return(main_nodes(a, opt));

	if (!strcmp(opt->cmd, "exit"))
		return(main_exit(a, opt));

	printf("Unknown subcommand: 'ben %s'. "
		"Try 'ben help' or 'ben man'.\n",
		opt->cmd);
	
	return(1);
}

// ****************************************************************************
// Main
// ****************************************************************************
int main(int argc, char **argv)
{
	args a;
	gopts opt;
	int r;
	
	if (exec_signals_setup())
		return 1;
	
	config_init();
	arg_init(&a, argc, argv);
	
	r = main_args(&a, &opt);
	
	arg_clear(&a);
	config_clear();
	
	return(r);
}

