/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "xmem.h"
#include "exec.h"
#include "network.h"


#define SOCKET_BACKLOG		10

// compat
#ifndef SOCK_CLOEXEC
#define SOCK_CLOEXEC		0
#endif


// **************************************************************************
// String utilities
// **************************************************************************
static int char_in(int c, char *list)
{
	while (*list) {
		if (c == *list)
			return(1);
		list++;
	}
	
	return(0);
}

static char *str_while(char *s, char *list)
{
	while (1) {
		if (s[0] == 0)
			return(s);
		if (!char_in(s[0], list))
			return(s);
		s++;
	}
}

static char *str_until(char *s, char *list)
{
	while (1) {
		if (s[0] == 0)
			return(s);
		if ((s[0] == '\\') && (s[1] != 0))
			s += 2;
		else if (char_in(s[0], list))
			return(s);
		else
			s++;
	}
}

static int str_copy(char *dst, size_t max, char *src, size_t len)
{
	size_t o, i;
	
	o = i = 0;
	
	while (i < len) {
		if ((src[i] == '\\') && (src[i + 1] != 0))
			i++;
		
		if (o + 1 >= max)
			return(-1);
		
		dst[o] = src[i];
		
		o++;
		i++;
	}
	
	dst[o] = 0;
	return(0);
}

// **************************************************************************
// Field parsing
// **************************************************************************
static int socket_field_parse(char *dst, size_t size, char **next, char *s)
{
	char *p, *q;
	
	// whitespace
	s = str_while(s, " \t\r\n");
	
	// bracket-escaped
	if (*s == '[') {
		s++;
		
		p = str_until(s, "]");
		
		if (*p != ']')
			return(-1);
		
		if (str_copy(dst, size, s, p - s))
			return(-1);
		
		q = str_while(p + 1, " \t\r\n");
		
		if ((*q != 0) && (*q != ':'))
			return(-1);

		*next = q;
		return(0);
	}
	
	// regular
	p = str_until(s, ":");
	
	if ((*p != 0) && (*p != ':'))
		return(-1);
	
	if (str_copy(dst, size, s, p - s))
		return(-1);
	
	*next = p;
	return(0);
}


// **************************************************************************
// Address parsing
// **************************************************************************
int socket_addr_parse(socket_addr *r, char *s)
{
	char *p, *q;
	
	if (socket_field_parse(r->path, SOCKET_PATH_LEN, &p, s))
		return(-1);
	
	// unix
	if (strchr(r->path, '/')) {
		r->type = SOCKET_UNIX;
		r->serv[0] = 0;
		return(0);
	}
	
	// inet
	r->type = SOCKET_INET;
	
	if (*p == ':') {
		if (socket_field_parse(r->serv, SOCKET_SERV_LEN, &q, p + 1))
			return(-1);
		
		if (*q != 0)
			return(-1);
	} else {
		r->serv[0] = 0;
	}

	return(0);
}

// **************************************************************************
// UNIX: path checking
// **************************************************************************
static int socket_unix_addr(struct sockaddr_un *d, char *path)
{
	size_t max;
	size_t sz;
	
	max = sizeof(((struct sockaddr_un *)0)->sun_path);
	sz = strlen(path) + 1;
	
	if (sz > max) {
		fprintf(stderr, "socket_unix_addr(): [%s]: path too long "
			"(%zd > %zd)\n", path, sz, max);
		return(-1);
	}
	
	if (d) {
		memset(d, 0, sizeof(*d));
		d->sun_family = AF_UNIX;
		memcpy(d->sun_path, path, sz);
	}
	
	return(0);
}

// **************************************************************************
// INET: DNS utility
// **************************************************************************
static struct addrinfo *socket_tcp_addrinfo(char *host, char *port, int server)
{
	struct addrinfo hints;
	struct addrinfo *res;
	int e;
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	hints.ai_flags = AI_ADDRCONFIG;
	
	if ((!host) || (host[0] == 0)
	||  ((host[0] == '*') && (host[1] == 0))) {
		host = NULL;
		if (server)
			hints.ai_flags |= AI_PASSIVE;
	} else if (!strcmp(host, "localhost")) {
		host = NULL;
	}
	
	if ((!port) || (port[0] == 0)) {
		fprintf(stderr, "socket_tcp_addrinfo(): port required\n");
		return(NULL);
	}
	
	e = getaddrinfo(host, port, &hints, &res);
	
	if (e != 0) {
		fprintf(stderr, "socket_tcp_addrinfo(): "
			"getaddrinfo(%s, %s): %s\n",
			(host) ? host : "*", port, gai_strerror(e));
		if (e == EAI_SYSTEM)
			fprintf(stderr, "\t(%s)\n", strerror(errno));
		return(NULL);
	}
	
	return(res);
}


// **************************************************************************
// UNIX bind
// **************************************************************************
static int socket_bind_unix(char *path)
{
	struct sockaddr_un sa;
	int fd;
	
	if (socket_unix_addr(&sa, path))
		return(-1);
	
	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	
	if (fd == -1) {
		fprintf(stderr, "socket_bind_unix(): socket(): %s\n",
			strerror(errno));
		return(-1);
	}
	
	if (unlink(path))
		errno = 0;
	
	if (exec_path_mkdir(path))
		errno = 0;
	
	if (bind(fd, (struct sockaddr *)&sa, sizeof(struct sockaddr_un))) {
		fprintf(stderr, "socket_bind_unix(): bind(%s): %s\n",
			path, strerror(errno));
		close(fd);
		unlink(path);
		return(-1);
	}
	
	if (listen(fd, SOCKET_BACKLOG)) {
		fprintf(stderr, "socket_bind_unix(): listen(%s): %s\n",
			path, strerror(errno));
		close(fd);
		unlink(path);
		return(-1);
	}
	
	
	fprintf(stderr, "socket: unix %s\n", path);
	return(fd);
}

// **************************************************************************
// INET bind
// **************************************************************************
static int socket_bind_tcp(char *host, char *port)
{
	socket_addr a;
	struct addrinfo *res, *rp;
	int fd;
	
	res = socket_tcp_addrinfo(host, port, 1);
	
	if (!res)
		return(-1);
	
	for (rp = res; rp != NULL; rp = rp->ai_next) {
		if (getnameinfo(rp->ai_addr, rp->ai_addrlen,
			a.path, sizeof(a.path), a.serv, sizeof(a.serv),
			NI_NUMERICHOST | NI_NUMERICSERV) == 0)
		{
			fprintf(stderr, "socket: tcp/ip %s %s\n",
				a.path, a.serv);
		}
		
		fd = socket(rp->ai_family, rp->ai_socktype | SOCK_CLOEXEC,
			rp->ai_protocol);
		
		if (fd == -1) {
			fprintf(stderr, "socket_bind_tcp(): socket(): %s\n",
				strerror(errno));
			errno = 0;
			continue;
		}
	
		if (bind(fd, rp->ai_addr, rp->ai_addrlen)) {
			fprintf(stderr, "socket_bind_tcp(): bind(): %s\n",
				strerror(errno));
			errno = 0;
			close(fd);
			continue;
		}
		
		if (listen(fd, SOCKET_BACKLOG)) {
			fprintf(stderr, "socket_bind_tcp(): listen(): %s\n",
				strerror(errno));
			errno = 0;
			close(fd);
			continue;
		}
		
		if (fcntl(fd, F_SETFL, O_NONBLOCK)) {
			fprintf(stderr, "socket_bind_tcp(): "
				"fcntl(F_SETFL, O_NONBLOCK): %s\n",
				strerror(errno));
			errno = 0;
			close(fd);
			continue;
		}
		
		break;
	}
	
	freeaddrinfo(res);
	
	if (!rp)
		return(-1);
	
	return(fd);
}

// **************************************************************************
// 
// **************************************************************************
int socket_bind(char *addr)
{
	socket_addr a;
	
	if (socket_addr_parse(&a, addr))
		return(-1);
	
	if (a.type == SOCKET_UNIX)
		return(socket_bind_unix(a.path));
	
	if (a.type == SOCKET_INET)
		return(socket_bind_tcp(a.path, a.serv));

	return(-1);
}


// **************************************************************************
// 
// **************************************************************************
int socket_unbind(int sfd, char *addr)
{
	socket_addr a;
	int r;
	
	if (socket_addr_parse(&a, addr))
		return(-1);
	
	r = 0;
	
	if (close(sfd)) {
		fprintf(stderr, "socket_unbind(): close(): %s\n",
			strerror(errno));
		r = -1;
	}
	
	if (a.type == SOCKET_UNIX) {
		if (unlink(a.path)) {
			fprintf(stderr, "socket_unbind(): unlink(%s): %s\n",
				a.path, strerror(errno));
			r = -1;
		}
	}
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int socket_accept(int sfd)
{
	socket_addr a;
	struct sockaddr_storage pa;
	socklen_t pa_len;
	int cfd, e;
	
	pa_len = sizeof(pa);
	cfd = accept(sfd, (struct sockaddr *)&pa, &pa_len);
	
	if (cfd == -1) {
		fprintf(stderr, "socket_accept(): accept(): %s\n",
			strerror(errno));
		return(-1);
	}
	
	if (fcntl(cfd, F_SETFL, O_NONBLOCK)) {
		fprintf(stderr, "socket_accept(): "
			"fcntl(F_SETFL, O_NONBLOCK): %s\n",
			strerror(errno));
		close(cfd);
		return(-1);
	}

	if ((pa.ss_family == AF_INET) || (pa.ss_family == AF_INET6)) {
		e = getnameinfo((struct sockaddr *)&pa, pa_len,
			a.path, sizeof(a.path), a.serv, sizeof(a.serv),
			NI_NUMERICHOST | NI_NUMERICSERV);
		
		if (e) {
			fprintf(stderr, "socket_accept(): getnameinfo(): %s\n",
				gai_strerror(e));
			if (e == EAI_SYSTEM)
				fprintf(stderr, "\t(%s)\n", strerror(errno));
			close(cfd);
			return(-1);
		}
		
		fprintf(stderr, "socket_accept(): host %s, port %s, fd %d",
			a.path, a.serv, cfd);
	}
	
	return(cfd);
}



// **************************************************************************
// 
// **************************************************************************
static int socket_connect_unix(char *path)
{
	struct sockaddr_un sa;
	int fd;
	
	if (socket_unix_addr(&sa, path))
		return(-1);
	
	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	
	if (fd == -1) {
		fprintf(stderr, "socket_connect_unix(): socket(): %s\n",
			strerror(errno));
		return(-1);
	}
	
	if (connect(fd, (struct sockaddr *)&sa, sizeof(struct sockaddr_un))) {
		fprintf(stderr, "socket_connect_unix(): connect(%s): %s\n",
			path, strerror(errno));
		close(fd);
		return(-1);
	}
	
	return(fd);
}

// **************************************************************************
// 
// **************************************************************************
static int socket_connect_tcp(char *host, char *port)
{
	socket_addr a;
	struct addrinfo *res, *rp;
	char *de, *dt;
	int fd, e;
	
	res = socket_tcp_addrinfo(host, port, 0);
	
	if (!res)
		return(-1);

	de = NULL;

	for (rp = res; rp != NULL; rp = rp->ai_next) {
		e = getnameinfo(rp->ai_addr, rp->ai_addrlen,
			a.path, sizeof(a.path), a.serv, sizeof(a.serv),
			NI_NUMERICHOST | NI_NUMERICSERV);
		if (e) {
			fprintf(stderr, "socket_connect(): getnameinfo(): %s\n",
				gai_strerror(e));
			if (e == EAI_SYSTEM)
				fprintf(stderr, "\t(%s)\n", strerror(errno));
			errno = 0;
			a.path[0] = '?';
			a.path[1] = 0;
			a.serv[0] = '?';
			a.serv[1] = 0;
		}

		//fprintf(stderr, "connect: tcp/ip %s %s\n", nethost, netserv);
	
		fd = socket(rp->ai_family, rp->ai_socktype | SOCK_CLOEXEC,
			rp->ai_protocol);
		
		if (fd == -1) {
			fprintf(stderr, "socket_connect(): socket(%s %s): %s\n",
				a.path, a.serv, strerror(errno));
			errno = 0;
			continue;
		}
	
		if (connect(fd, rp->ai_addr, rp->ai_addrlen)) {
			dt = xprintf(NULL,
				"%ssocket_connect(): connect(%s %s): %s\n",
				(de) ? de : "",
				a.path, a.serv, strerror(errno));
			errno = 0;
			
			if (de)
				free(de);
			de = dt;
			
			close(fd);
			continue;
		}
		
		if (fcntl(fd, F_SETFL, O_NONBLOCK)) {
			fprintf(stderr, "socket_connect(): "
				"fcntl(F_SETFL, O_NONBLOCK): %s\n",
				strerror(errno));
			errno = 0;
			close(fd);
			continue;
		}
		
		break;
	}
	
	freeaddrinfo(res);
	
	if (!rp) {
		if (de) {
			fprintf(stderr, "%s", de);
			free(de);
		}
		return(-1);
	}
	
	// silently ignore delayed errors if successful
	if (de)
		free(de);
	
	return(fd);
}

// **************************************************************************
// 
// **************************************************************************
int socket_connect(char *addr)
{
	socket_addr a;
	
	if (socket_addr_parse(&a, addr))
		return(-1);
	
	if (a.type == SOCKET_UNIX)
		return(socket_connect_unix(a.path));
	
	if (a.type == SOCKET_INET)
		return(socket_connect_tcp(a.path, a.serv));

	return(-1);
}


