/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2023  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include "xmem.h"
#include "buffer.h"
#include "pattern.h"
#include "queue.h"
#include "node.h"

// ***************************************************************************
// 
// ***************************************************************************
node *node_new(int id, int fd)
{
	node *b;
	
	b = xmalloc(sizeof(node));
	
	b->fd = fd;
	b->id = id;
	b->name = xstrdup("_");
	b->processes = -1;
	b->flags = 0;
	b->refs = 0;
	b->atime = gtime_zero();
	b->ping_sent = 0;
	b->ping_received = 0;
	b->ack_h = -1;
	b->ack_n = 0;
	
	buffer_init(&b->inbuff, 0);
	buffer_init(&b->outbuff, 0);
	
	queue_init(&b->running);
	
	b->prev = NULL;
	b->next = NULL;
	
	return(b);
}


// ***************************************************************************
// 
// ***************************************************************************
void node_free(node *b)
{
	xfree(b->name);
	
	buffer_clear(&b->inbuff);
	buffer_clear(&b->outbuff);
	queue_clear(&b->running);
	
	xfree(b);
}

// ***************************************************************************
// 
// ***************************************************************************
void node_setup(node *b, char *name, int cores)
{
	if (b->name)
		xfree(b->name);
	
	b->name = xstrdup(name);
	b->processes = cores;
}


// ***************************************************************************
// 
// ***************************************************************************
int node_matches(node *b, pattern *pat)
{
	return pattern_matches(pat, b->id, b->name);
}

// ***************************************************************************
// 
// ***************************************************************************
void nodeset_init(nodeset *bs)
{
	bs->head = NULL;
	bs->tail = NULL;
	bs->n = 0;
}

void nodeset_clear(nodeset *bs)
{
	node *b, *c;
	
	b = bs->head;
	while (b) {
		c = b->next;
		
		node_free(b);
		
		b = c;
	}
	
	bs->head = NULL;
	bs->tail = NULL;
	bs->n = 0;
}

// ***************************************************************************
// 
// ***************************************************************************
void nodeset_attach(nodeset *bs, node *b)
{
	b->prev = bs->tail;
	b->next = NULL;
	
	if (bs->tail)
		bs->tail->next = b;
	
	bs->tail = b;
	
	if (!bs->head)
		bs->head = b;
	
	bs->n++;
}

void nodeset_detach(nodeset *bs, node *b)
{
	if (b->prev)
		b->prev->next = b->next;
	
	if (b->next)
		b->next->prev = b->prev;
	
	if (bs->head == b)
		bs->head = b->next;
	
	if (bs->tail == b)
		bs->tail = b->prev;
	
	b->prev = NULL;
	b->next = NULL;
	
	bs->n--;
}

// ***************************************************************************
// 
// ***************************************************************************
node *nodeset_find(nodeset *bs, int id)
{
	node *b = bs->head;
	
	while (b) {
		if (b->id == id)
			return(b);
		
		b = b->next;
	}
	
	return(NULL);
}


